

package WebFlow;
import WebFlow.event.*;

import java.util.*;
import java.util.Map.Entry;

public class      BeanContextServicesSupport extends BeanContextSupport
       implements BeanContextServicesOperations {

    /**
     * <p>
     * Construct a BeanContextServicesSupport instance
     * </p>
     *
     * @param peer	The peer BeanContext we are supplying an implementation for, if null the this object is its own peer
     * 
     */

    public BeanContextServicesSupport(BeanContextServices peer, 
									  org.omg.CORBA.ORB orb)
		throws NullPointerException
	{
	    super(peer, orb);
    }
    
	public boolean isEqual(org.omg.CORBA.Object o1, org.omg.CORBA.Object o2)
	{
	    return (orb.object_to_string(o1)).equals(orb.object_to_string(o2));
    }

    /**
     * called by BeanContextSupport superclass during construction and 
     * deserialization to initialize subclass transient state.
     * subclasses may envelope this method, but should not override it or
     * call it directly.
     */

    public void initialize() {
		super.initialize();

		services     = new myHashMap(serializable + 1);
		bcsListeners = new ArrayList(1);
    }

    /**
     * @return the instance of BeanContext this object is providing the implemen
tation for.
     */
 
    public BeanContextServices getBeanContextServicesPeer() { 
		return (BeanContextServices)getBeanContextChildPeer();
    }

    /************************************************************************/

    /*
     * protected nested class containing per child information, an instance
     * of which is associated with each child in the "children" hashtable.
     * subclasses can extend this class to include their own per-child state.
     *
     * Note that this 'value' is serialized with the corresponding child 'key'
     * when the BeanContextSupport is serialized.
     */

    protected class BCSSChild extends BeanContextSupport.BCSChild  {

	/*
	 * private nested class to map serviceClass to Provider and requestors
	 * listeners.
	 */

	class BCSSCServiceClassRef {

	    // create an instance of a service ref

	    BCSSCServiceClassRef(String sc, BeanContextServiceProvider bcsp, boolean delegated) {
		super();
	
		serviceClass     = sc;

		if (delegated)
		    delegateProvider = bcsp;
		else 
		    serviceProvider  = bcsp;
	    }

	    // add a requestor and assoc listener

	    void addRequestor(org.omg.CORBA.Object requestor, BeanContextServiceRevokedListener bcsrl)
		 throws TooManyListenersException {

		BeanContextServiceRevokedListener cbcsrl = 
		 (BeanContextServiceRevokedListener)requestors.get(requestor);
	
		//if (cbcsrl != null && !cbcsrl.equals(bcsrl))
		if (cbcsrl != null && ! isEqual(cbcsrl,bcsrl))
		    throw new TooManyListenersException();

		requestors.put(requestor, bcsrl);
	    }

	    // remove a requestor

	    void removeRequestor(Object requestor) {
		requestors.remove(requestor);
	    }

	    // check a requestors listener

	    void verifyRequestor(Object requestor, BeanContextServiceRevokedListener bcsrl) throws TooManyListenersException {
		BeanContextServiceRevokedListener cbcsrl = 
		 (BeanContextServiceRevokedListener)requestors.get(requestor);

		//if (cbcsrl != null && !cbcsrl.equals(bcsrl))
		if (cbcsrl != null && ! isEqual(cbcsrl,bcsrl))
		    throw new TooManyListenersException();
	    }

	    void verifyAndMaybeSetProvider(BeanContextServiceProvider bcsp, 
									   boolean isDelegated)
									   throws UnsupportedOperationException
		{
		BeanContextServiceProvider current;

		if (isDelegated) { // the provider is delegated
		    current = delegateProvider;

		    if (current == null || bcsp == null) {
			delegateProvider = bcsp;
			return;
		    }
		} else { // the provider is registered with this BCS
		    current = serviceProvider;

		    if (current == null || bcsp == null) {
			serviceProvider = bcsp;
			return;
		    }
		}

		//if (!current.equals(bcsp))
		if (!isEqual(current, bcsp))
		    throw new UnsupportedOperationException("existing service reference obtained from different BeanContextServiceProvider not supported");

	    }

	    java.util.Iterator entries() { return requestors.entrySet().iterator(); }

	    boolean isEmpty() { return requestors.isEmpty(); }

	    String getServiceClass() { return serviceClass; }

	    BeanContextServiceProvider getServiceProvider() {
		return serviceProvider;
	    }

	    BeanContextServiceProvider getDelegateProvider() {
		return delegateProvider;
	    }

	    boolean isDelegated() { return delegateProvider != null; }

 	    void addRef(boolean delegated) { 
		if (delegated)
		    delegateRefs++;
		else
		    serviceRefs++;
	    }


	    void releaseRef(boolean delegated) { 
                if (delegated)
                    if (--delegateRefs == 0) delegateProvider = null;
                else
                    if (--serviceRefs  <= 0) serviceProvider = null;
            }    

	    int getRefs() { return serviceRefs + delegateRefs; }

	    int getDelegateRefs() { return delegateRefs; }

	    int getServiceRefs() { return serviceRefs; }

	    /*
	     * fields
	     */

	    String				serviceClass;

	    BeanContextServiceProvider		serviceProvider;
	    int 				serviceRefs;

	    BeanContextServiceProvider		delegateProvider; // proxy
	    int								delegateRefs;

	    myHashMap				requestors = new myHashMap(1);
	}

     	/*
	 * per service reference info ...
	 */

	class BCSSCServiceRef {
	    BCSSCServiceRef(BCSSCServiceClassRef scref, boolean isDelegated) {
		serviceClassRef = scref;
		delegated       = isDelegated;
	    }

	    void addRef()  { refCnt++;        }
	    int  release() { return --refCnt; }

	    BCSSCServiceClassRef getServiceClassRef() { return serviceClassRef; }

	    boolean isDelegated() { return delegated; }

	    /*
    	     * fields
	     */

	    BCSSCServiceClassRef serviceClassRef;
	    int			 refCnt	   = 1;
	    boolean		 delegated = false;
	}

	BCSSChild(org.omg.CORBA.Object bcc, 
	org.omg.CORBA.Object peer) 
	{ 
		super(bcc, peer);
	}

        // note usage of service per requestor, per service

	synchronized void usingService(org.omg.CORBA.Object requestor,
		org.omg.CORBA.Object service, 
		String serviceClass,
		BeanContextServiceProvider bcsp,
		boolean isDelegated,
		BeanContextServiceRevokedListener bcsrl) 
		throws TooManyListenersException, UnsupportedOperationException
		
	{
	
	    // first, process mapping from serviceClass to requestor(s)

	    BCSSCServiceClassRef serviceClassRef = null;

	    if (serviceClasses == null) 
		serviceClasses = new myHashMap(1);
	    else
	        serviceClassRef = (BCSSCServiceClassRef)serviceClasses.get(serviceClass);

	    if (serviceClassRef == null) { // new service being used ...
	    	serviceClassRef = new BCSSCServiceClassRef(serviceClass, bcsp, isDelegated);
		serviceClasses.put(serviceClass, serviceClassRef);

		serviceClassRef.addRequestor(requestor, bcsrl);

	    } else { // existing service ...
		serviceClassRef.verifyAndMaybeSetProvider(bcsp, isDelegated); // throws
		serviceClassRef.verifyRequestor(requestor, bcsrl); // throws
	    }

	    serviceClassRef.addRef(isDelegated);

	    // now handle mapping from requestor to service(s)

	    BCSSCServiceRef serviceRef = null;
	    Map		    services   = null;

	    if (serviceRequestors == null) {
		serviceRequestors = new myHashMap(1);
	    } else {
		services = (Map)serviceRequestors.get(requestor);
	    }

	    if (services == null) {
		services = new myHashMap(1);

		serviceRequestors.put(requestor, services);
	    } else 
		serviceRef = (BCSSCServiceRef)services.get(service);

	    if (serviceRef == null) {
		serviceRef = new BCSSCServiceRef(serviceClassRef, isDelegated);

		services.put(service, serviceRef);
	    } else {
		serviceRef.addRef();
	    }
	}

	// release a service reference

	synchronized void releaseService(org.omg.CORBA.Object requestor,
									 org.omg.CORBA.Object service)
	{
	    if (serviceRequestors == null) return;

	    Map services = (Map)serviceRequestors.get(requestor);
	
	    if (services == null) return; // oops its not there anymore!

	    BCSSCServiceRef serviceRef = (BCSSCServiceRef)services.get(service);

	    if (serviceRef == null) return; // oops its not there anymore!
	    
	    BCSSCServiceClassRef serviceClassRef = serviceRef.getServiceClassRef();
	    boolean		       isDelegated = serviceRef.isDelegated();
	    BeanContextServiceProvider bcsp        = isDelegated ? serviceClassRef.getDelegateProvider() : serviceClassRef.getServiceProvider();

	    bcsp.releaseService(BeanContextServicesSupport.this.getBeanContextServicesPeer(), requestor, service);

	    serviceClassRef.releaseRef(isDelegated);

	    if (serviceRef.release() == 0) {

		services.remove(service);

		if (services.isEmpty()) {
		    serviceRequestors.remove(requestor);
		    serviceClassRef.removeRequestor(requestor);
		}

		if (serviceRequestors.isEmpty()) {
		    serviceRequestors = null;
		}

		if (serviceClassRef.isEmpty()) {
		    serviceClasses.remove(serviceClassRef.getServiceClass());
		}

		if (serviceClasses.isEmpty())
		    serviceClasses = null;
	    }
	}

	// revoke a service

	synchronized void revokeService(String serviceClass, boolean isDelegated,
	 boolean revokeNow) throws IllegalArgumentException
	 {
	    if (serviceClasses == null) return;

	    BCSSCServiceClassRef serviceClassRef = (BCSSCServiceClassRef)serviceClasses.get(serviceClass);

	    if (serviceClassRef == null) return;

	    java.util.Iterator i = serviceClassRef.entries();

		BeanContextServiceRevokedEventImpl bcsre;
		bcsre = new BeanContextServiceRevokedEventImpl(BeanContextServicesSupport.this.getBeanContextServicesPeer(), serviceClass, revokeNow);
			    
		boolean			   noMoreRefs  = false;

	    while (i.hasNext()) {
		Map.Entry	    		  entry    = (Map.Entry)i.next();
		BeanContextServiceRevokedListener listener = (BeanContextServiceRevokedListener)entry.getValue();

		if (revokeNow) {
		    Object  requestor = entry.getKey();
		    Map     services  = (Map)serviceRequestors.get(requestor);
		    if (services != null) {
			java.util.Iterator i1 = services.entrySet().iterator();

			while (i1.hasNext()) {
			    Map.Entry       tmp        = (Map.Entry)i1.next();

			    BCSSCServiceRef serviceRef = (BCSSCServiceRef)tmp.getValue();
			    if (serviceRef.getServiceClassRef().equals(serviceClassRef) && isDelegated == serviceRef.isDelegated()) {	
				   i1.remove();
			    }
			}

			if (noMoreRefs = services.isEmpty()) {
			    serviceRequestors.remove(requestor);
			}
		    }

		    if (noMoreRefs) serviceClassRef.removeRequestor(requestor);
		}

		 _BeanContextServiceRevokedEventImplBase_tie o_bcsre;
		  o_bcsre = new _BeanContextServiceRevokedEventImplBase_tie(bcsre);
		  listener.serviceRevoked(o_bcsre);
	    }

	    if (revokeNow && serviceClasses != null) {
		if (serviceClassRef.isEmpty())
		    serviceClasses.remove(serviceClass);

	        if (serviceClasses.isEmpty())
		    serviceClasses = null;
	    }

	    if (serviceRequestors != null && serviceRequestors.isEmpty())
		serviceRequestors = null;
	}

	// release all references for this child since it has been unnested.

	void cleanupReferences() {

	    if (serviceRequestors == null) return;

	    java.util.Iterator requestors = serviceRequestors.entrySet().iterator();

	    while(requestors.hasNext()) {
		Map.Entry 	     tmp       = (Map.Entry)requestors.next();

		//Because key is IOR (stringfied version of CORBA Object),
		//we need to transform it to real object
		//org.omg.CORBA.Object    requestor = (org.omg.CORBA.Object)tmp.getKey();
		String s = (String)tmp.getKey();
		org.omg.CORBA.Object    requestor = orb.string_to_object(s);
		 
		java.util.Iterator 	     services  = ((Map)tmp.getValue()).entrySet().iterator();

		requestors.remove();

		while (services.hasNext()) {
		    Map.Entry       entry   = (Map.Entry)services.next();
		    org.omg.CORBA.Object  service = (org.omg.CORBA.Object)entry.getKey();
		    BCSSCServiceRef sref    = (BCSSCServiceRef)entry.getValue();

		    BCSSCServiceClassRef       scref = sref.getServiceClassRef();
		
		    BeanContextServiceProvider bcsp  = sref.isDelegated() ? scref.getDelegateProvider() : scref.getServiceProvider();

		    scref.removeRequestor(requestor);
		    services.remove();

		    while (sref.release() >= 0) {
			bcsp.releaseService(BeanContextServicesSupport.this.getBeanContextServicesPeer(), requestor, service);
		    }
		}
	    }

	    serviceRequestors = null;
	    serviceClasses    = null;
	}

        void revokeAllDelegatedServicesNow()
		throws IllegalArgumentException
        {
	    if (serviceClasses == null) return;

	    java.util.Iterator serviceClassRefs  = serviceClasses.values().iterator();

	    while (serviceClassRefs.hasNext()) {
	        BCSSCServiceClassRef serviceClassRef = (BCSSCServiceClassRef)serviceClassRefs.next();

	    	if (!serviceClassRef.isDelegated()) continue;

	        java.util.Iterator 		       i           = serviceClassRef.entries();
	    	BeanContextServiceRevokedEventImpl bcsre       = new BeanContextServiceRevokedEventImpl(BeanContextServicesSupport.this.getBeanContextServicesPeer(), serviceClassRef.getServiceClass(), true);
	    	boolean			       noMoreRefs  = false;

	    	while (i.hasNext()) {
		    Map.Entry	    		      entry     = (Map.Entry)i.next();
		    BeanContextServiceRevokedListener listener  = (BeanContextServiceRevokedListener)entry.getValue();

		    Object     requestor = entry.getKey();
		    Map       services  = (Map)serviceRequestors.get(requestor);

		    if (services != null) {
			java.util.Iterator i1 = services.entrySet().iterator();

			while (i1.hasNext()) {
			    Map.Entry       tmp        = (Map.Entry)i1.next();

			    BCSSCServiceRef serviceRef = (BCSSCServiceRef)tmp.getValue();
			    if (serviceRef.getServiceClassRef().equals(serviceClassRef) && serviceRef.isDelegated()) {	
				i1.remove();
			    }
			}

			if (noMoreRefs = services.isEmpty()) {
			    serviceRequestors.remove(requestor);
			}
		    }

		    if (noMoreRefs) serviceClassRef.removeRequestor(requestor);

			_BeanContextServiceRevokedEventImplBase_tie o_bcsre;
			o_bcsre = new _BeanContextServiceRevokedEventImplBase_tie(bcsre);	 
		    listener.serviceRevoked(o_bcsre);

		    if (serviceClassRef.isEmpty())
		        serviceClasses.remove(serviceClassRef.getServiceClass());
	    	}
	    }

	    if (serviceClasses.isEmpty()) serviceClasses = null;

	    if (serviceRequestors != null && serviceRequestors.isEmpty())
		serviceRequestors = null;
        }

	/*
 	 * fields 
	 */

	private transient myHashMap	serviceClasses;	   
	private transient myHashMap	serviceRequestors;
    }

    /**
     * <p>
     * Subclasses can override this method to insert their own subclass
     * of Child without having to override add() or the other Collection
     * methods that add children to the set.
     * </p>
     *
     * @param targetChild the child to create the Child on behalf of
     * @param peer        the peer if the targetChild and peer are related by BeanContextProxy
     */

    protected BCSChild createBCSChild(org.omg.CORBA.Object targetChild,
					 org.omg.CORBA.Object peer)
	{
		return new BCSSChild(targetChild, peer);
    }

    /************************************************************************/

	/**
	 * subclasses may subclass this nested class to add behaviors for
	 * each BeanContextServicesProvider.
	 */

	protected static class BCSSServiceProvider{

	    BCSSServiceProvider(String sc, BeanContextServiceProvider bcsp) {
		super();
	
		serviceProvider = bcsp;
	    }

	    protected BeanContextServiceProvider getServiceProvider() {
		return serviceProvider;
	    }

	    /*
   	     * fields
	     */

	    protected BeanContextServiceProvider serviceProvider;
	}

    	/**
	 * subclasses can override this method to create new subclasses of
	 * BCSSServiceProvider without having to overrride addService() in
	 * order to instantiate.
	 */

	protected BCSSServiceProvider createBCSSServiceProvider(String sc, BeanContextServiceProvider bcsp) {
	    return new BCSSServiceProvider(sc, bcsp);
	}

    /************************************************************************/

    /**
     * add a BeanContextServicesListener
     *
     * @throw new NullPointerException
     */

    public void addBeanContextServicesListener(BeanContextServicesListener bcsl) 
		throws NullPointerException
	{
	if (bcsl == null) throw new NullPointerException("bcsl");

	synchronized(bcsListeners) {
	    if (bcsListeners.contains(bcsl))
		return;
	    else
	        bcsListeners.add(bcsl);
	}
    }

    /**
     * remove a BeanContextServicesListener
     */

    public void removeBeanContextServicesListener(BeanContextServicesListener bcsl) 
		throws NullPointerException 
	{
	if (bcsl == null) throw new NullPointerException("bcsl");

	synchronized(bcsListeners) {
	    if (!bcsListeners.contains(bcsl))
		return;
	    else
	        bcsListeners.remove(bcsl);
	}
    }

    /**
     * add a service
     */

    public boolean addService(String serviceClass, BeanContextServiceProvider bcsp) 
	{
		try{
			return addService(serviceClass, bcsp, true);
		}catch(NullPointerException e1){
			return false;
		 }
		 catch(IllegalArgumentException e2){
			return false;
		 };
    }

    /**
     * add a service
     */

    protected boolean addService(String serviceClass, 
								BeanContextServiceProvider bcsp, 
								boolean fireEvent) 
			throws NullPointerException,IllegalArgumentException
	{

	if (serviceClass == null) throw new NullPointerException("serviceClass");
	if (bcsp         == null) throw new NullPointerException("bcsp");

	            if (services.containsKey(serviceClass))
    	        return false;
    	    else {
    	        services.put(serviceClass,  createBCSSServiceProvider(serviceClass, bcsp));
    
				/* TO DO LIST */
    	        //if (bcsp instanceof Serializable) serializable++;
    
    	        if (!fireEvent) return true;
    
    
    	        BeanContextServiceAvailableEventImpl bcssae = new BeanContextServiceAvailableEventImpl(getBeanContextServicesPeer(), serviceClass);
				_BeanContextServiceAvailableEventImplBase_tie o_bcssae;
				o_bcssae = new _BeanContextServiceAvailableEventImplBase_tie(bcssae);

    	        fireServiceAdded(o_bcssae);
    
    	        synchronized(children) {
    	            java.util.Iterator i = children.keySet().iterator();
    
    	            while (i.hasNext()) {
    	                //Object c = i.next();
						org.omg.CORBA.Object c = (org.omg.CORBA.Object)i.next();
    	                if (c instanceof BeanContextServices) {
    		            //((BeanContextServicesListener)c).serviceAvailable(o_bcssae);
    					  (BeanContextServicesListenerHelper.narrow(c)).serviceAvailable(o_bcssae);	
					    }
    	            }
    	        }
    
    	        return true;
    	    }
    }

    /**
     * remove a service
     */

    public void revokeService(String serviceClass, 
						      BeanContextServiceProvider bcsp, 
							  boolean revokeCurrentServicesNow)
		  throws NullPointerException,IllegalArgumentException
   {

	if (serviceClass == null) throw new NullPointerException("serviceClass");
	if (bcsp         == null) throw new NullPointerException("bcsp");

	    if (!services.containsKey(serviceClass)) return;

	    BCSSServiceProvider bcsssp = (BCSSServiceProvider)services.get(serviceClass);
	
	    if (!bcsssp.getServiceProvider().equals(bcsp)) 
	        throw new IllegalArgumentException("service provider mismatch");

	    services.remove(serviceClass);
			
			/* TO DO LIST */
            //if (bcsp instanceof Serializable) serializable--;

	    java.util.Iterator i = bcsChildren(); // get the BCSChild values.

	    while (i.hasNext()) {
	        ((BCSSChild)i.next()).revokeService(serviceClass, false, revokeCurrentServicesNow);
	    }

	    fireServiceRevoked(serviceClass, revokeCurrentServicesNow);

    }

    /**
     * has a service, which may be delegated
     */

    public synchronized boolean hasService(String serviceClass)
	throws NullPointerException
	{
	    if (serviceClass == null) throw new NullPointerException("serviceClass");

	    if (services.containsKey(serviceClass)) return true;

	    BeanContextServices bcs = null;

	    try {
	        bcs = (BeanContextServices)getBeanContext();
	    } catch (ClassCastException cce) {
	        return false;
	    }

	    return bcs == null ? false : bcs.hasService(serviceClass);
    }

    /************************************************************************/

    /*
     * a nested subclass used to represent a proxy for serviceClasses delegated
     * to an enclosing BeanContext.
     */

    protected class BCSSProxyServiceProvider implements 
								   BeanContextServiceProviderOperations,
							       BeanContextServiceRevokedListenerOperations
	 {
        BCSSProxyServiceProvider(BeanContextServices bcs) 
		{
			super();
			nestingCtxt = bcs;
		}

	public org.omg.CORBA.Object getService(BeanContextServices bcs, 
							 org.omg.CORBA.Object requestor, 
							 String serviceClass, 
							 org.omg.CORBA.Object serviceSelector)
   {
	    org.omg.CORBA.Object service = null;

		_BeanContextServiceRevokedListenerImplBase_tie o_this;
		o_this = new _BeanContextServiceRevokedListenerImplBase_tie(this);
					
	    try { 
	 		service = nestingCtxt.getService(bcs, requestor, serviceClass, 
											 serviceSelector, o_this);
	    } catch (TooManyListenersException tmle) {
			return null;
	    }
		catch  (UnsupportedOperationException e){
			return null;
	    }
        catch  (NullPointerException e){
			return null;
	    }
		catch  (IllegalArgumentException e){
			return null;
	    }


	
	    return service;
	}

	public void releaseService(BeanContextServices bcs, 
							   org.omg.CORBA.Object requestor, 
							   org.omg.CORBA.Object service)
    {
		try{
			nestingCtxt.releaseService(bcs, requestor, service);
		}catch(NullPointerException e){
			System.out.println("releaseService :"+e);
		}catch(IllegalArgumentException e){
			System.out.println("releaseService :"+e);
		};
	}

	public Iterator getCurrentServiceSelectors(BeanContextServices bcs, 
											   String serviceClass)
	 {
	    return nestingCtxt.getCurrentServiceSelectors(serviceClass);
	}

	public void serviceRevoked(BeanContextServiceRevokedEvent bcsre) 
	throws IllegalArgumentException
	{
	    java.util.Iterator i = bcsChildren(); // get the BCSChild values.

	    while (i.hasNext()) {
	        ((BCSSChild)i.next()).revokeService(bcsre.getServiceClass(), true, bcsre.isCurrentServiceInvalidNow());
	    }
	}

	/*
	 * fields
	 */

	private BeanContextServices nestingCtxt;
    }

    /************************************************************************/

    /**
     * obtain a service which may be delegated
     */

     public org.omg.CORBA.Object getService(BeanContextChild child, 
				org.omg.CORBA.Object requestor, 
				String serviceClass, 
				org.omg.CORBA.Object serviceSelector, 
				BeanContextServiceRevokedListener bcsrl) 
				throws TooManyListenersException, 
				UnsupportedOperationException,
				NullPointerException,
				IllegalArgumentException
	{
	
	if (child        == null) throw new NullPointerException("child");
	if (serviceClass == null) throw new NullPointerException("serviceClass");
	if (requestor    == null) throw new NullPointerException("requestor");
	if (bcsrl        == null) throw new NullPointerException("bcsrl");

	org.omg.CORBA.Object   service = null;
	BCSSChild              bcsc;
	BeanContextServices bcssp   = getBeanContextServicesPeer();

	    synchronized(children) { bcsc = (BCSSChild)children.get(child); }

	    if (bcsc == null) throw new IllegalArgumentException("not a child of this context"); // not a child ...

	    BCSSServiceProvider bcsssp = (BCSSServiceProvider)services.get(serviceClass);

	    if (bcsssp != null) {
	        BeanContextServiceProvider bcsp = bcsssp.getServiceProvider();
	        service = bcsp.getService(bcssp, requestor, serviceClass, serviceSelector);
	        if (service != null) { // do bookkeeping ...
		    try {
	                bcsc.usingService(requestor, service, serviceClass, bcsp, false, bcsrl);
		    } catch (TooManyListenersException tmle) {
		        bcsp.releaseService(bcssp, requestor, service);
		        throw tmle;
		    } catch (UnsupportedOperationException uope) {
		        bcsp.releaseService(bcssp, requestor, service);
		        throw uope; // unchecked rt exception
		    }

		    return service;
	        }
	    }


	    if (proxy != null) {

	        // try to delegate ...

	        service = proxy.getService(bcssp, requestor, serviceClass, serviceSelector);
	
	        if (service != null) { // do bookkeeping ...
		    /*try { */
			_BeanContextServiceProviderImplBase_tie o_proxy;
			o_proxy = new _BeanContextServiceProviderImplBase_tie(proxy);
		    bcsc.usingService(requestor, service, serviceClass, o_proxy, true, bcsrl);
		    
			/*} catch (TooManyListenersException tmle) {
		        proxy.releaseService(bcssp, requestor, service);
		        throw tmle;
		    } 
			catch (UnsupportedOperationException uope) {
		        proxy.releaseService(bcssp, requestor, service);
		        throw uope; // unchecked rt exception
		    }
			*/

		    return service;
	        }
	    }
	//}

       	return null;
    }

    /**
     * release a service
     */

    public void releaseService(BeanContextChild child, 
	  org.omg.CORBA.Object requestor, org.omg.CORBA.Object service) 
	  throws NullPointerException,IllegalArgumentException
	{
	if (child     == null) throw new NullPointerException("child");
	if (requestor == null) throw new NullPointerException("requestor");
	if (service   == null) throw new NullPointerException("service");

	BCSSChild bcsc;

	        synchronized(children) { bcsc = (BCSSChild)children.get(child); }

	        if (bcsc != null)
		    bcsc.releaseService(requestor, service);
		else
		   throw new IllegalArgumentException("child actual is not a child of this BeanContext");
     }

    /**
     * @return an iterator for all the currently registered service classes.
     */

    public Iterator getCurrentServiceClasses() {
	return new BCSIterator(services.keySet().iterator());
    }

    /**
     * @return an iterator for all the currently available service selectors
     * (if any) available for the specified service.
     */

    public WebFlow.Iterator getCurrentServiceSelectors(String serviceClass) {
	
		BCSSServiceProvider bcsssp = (BCSSServiceProvider)services.get(serviceClass);
		return null;
		/*
		return bcsssp != null ? 
		new BCSIterator(bcsssp.getServiceProvider().getCurrentServiceSelectors(getBeanContextServicesPeer(), serviceClass)) : null;
		*/
    }

    /**
     * BeanContextServicesListener callback, propagates event to all 
     * currently registered listeners and BeanContextServices children,
     * if this BeanContextService does not already implement this service
     * itself.
     *
     * subclasses may override or envelope this method to implement their
     * own propagation semantics.
     */

     public void serviceAvailable(BeanContextServiceAvailableEvent bcssae)
		 throws IllegalArgumentException
	  {
	    if (services.containsKey(bcssae.getServiceClass())) return;

	    fireServiceAdded(bcssae);

	    java.util.Iterator i;

	    synchronized(children) {
	        i = children.keySet().iterator();
	    }

	    while (i.hasNext()) {
			//Object c = i.next();
	        org.omg.CORBA.Object c = (org.omg.CORBA.Object)i.next();

	        //if (c instanceof BeanContextServices) {
			if (c._is_a(_BeanContextServicesImplBase._ob_ids_[0])){
	            //((BeanContextServicesListener)c).serviceAvailable(bcssae);
				(BeanContextServicesListenerHelper.narrow(c)).serviceAvailable(bcssae);
	        }
	    }
     }

    /**
     * BeanContextServicesListener callback, propagates event to all 
     * currently registered listeners and BeanContextServices children,
     * if this BeanContextService does not already implement this service
     * itself.
     *
     * subclasses may override or envelope this method to implement their
     * own propagation semantics.
     */

    public void serviceRevoked(BeanContextServiceRevokedEvent bcssre) 
		throws IllegalArgumentException
	{
	    if (services.containsKey(bcssre.getServiceClass())) return;

	    fireServiceRevoked(bcssre);

	    java.util.Iterator i;

	    synchronized(children) {
	        i = children.keySet().iterator();
	    }

	    while (i.hasNext()) {
	        //Object c = i.next();
			org.omg.CORBA.Object c = (org.omg.CORBA.Object)i.next();

	        //if (c instanceof BeanContextServices) {
			if (c._is_a(_BeanContextServicesImplBase._ob_ids_[0])){
	            //((BeanContextServicesListener)c).serviceRevoked(bcssre);
				(BeanContextServicesListenerHelper.narrow(c)).serviceRevoked(bcssre);
	        }
	    }
    }

    /**
     * @return the BeanContextServicesListener (if any) of the specified child
     */

    protected static final BeanContextServicesListener getChildBeanContextServicesListener(Object child) {
	try {
	    return (BeanContextServicesListener)child;
	} catch (ClassCastException cce) {
	    return null;
	}
    }

    /**
     * called from superclass child removal operations after a child
     * has been successfully removed. called with child synchronized.
     *
     * This subclass uses this hook to immediately revoke any services
     * being used by this child if it is a BeanContextChild.
     *
     * subclasses may envelope this method in order to implement their
     * own child removal side-effects.
     */

    protected void childJustRemovedHook(Object child, BCSChild bcsc) {
	BCSSChild bcssc = (BCSSChild)bcsc;

	bcssc.cleanupReferences();
    }

    /**
     * called from setBeanContext to notify a BeanContextChild
     * to release resources obtained from the nesting BeanContext.
     * 
     * This method revokes any services obtained from its parent.
     * 
     * subclasses may envelope this method to implement their own semantics.
     */

    protected synchronized void releaseBeanContextResources() {
	Object[] bcssc;

	super.releaseBeanContextResources();

	synchronized(children) {
	    if (children.isEmpty()) return;

	    bcssc = children.values().toArray();
	}


	for (int i = 0; i < bcssc.length; i++) {
		try{
			((BCSSChild)bcssc[i]).revokeAllDelegatedServicesNow();
		}catch(IllegalArgumentException e){
			System.out.println("releaseBeanContextResources :"+e);
		}

	}

	proxy = null;
    }

    /**
     * called from setBeanContext to notify a BeanContextChild
     * to allocate resources obtained from the nesting BeanContext.
     * 
     * subclasses may envelope this method to implement their own semantics.
     */

    protected synchronized void initializeBeanContextResources() {
	super.initializeBeanContextResources();
	//Inittialize orb
	orb = org.omg.CORBA.ORB.init();
	
	//BeanContext nbc = getBeanContext();
    org.omg.CORBA.Object  nbc = getBeanContext();

	if (nbc == null) return;

	try {
	    BeanContextServices bcs = (BeanContextServices)nbc;
	    proxy = new BCSSProxyServiceProvider(bcs);
	} catch (ClassCastException cce) {
	    // do nothing ...
	}
    }

    /**
     * Fire a BeanContextServiceEvent notifying of a new service
     */

    protected final void fireServiceAdded(String serviceClass) 
      throws IllegalArgumentException
	{
		BeanContextServiceAvailableEventImpl bcssae = new BeanContextServiceAvailableEventImpl(getBeanContextServicesPeer(), serviceClass); 
		_BeanContextServiceAvailableEventImplBase_tie o_bcssae;
		o_bcssae = new _BeanContextServiceAvailableEventImplBase_tie(bcssae);	
		fireServiceAdded(o_bcssae);
    } 

    protected final void fireServiceAdded(BeanContextServiceAvailableEvent bcssae) 
	throws IllegalArgumentException
	{
		Object[]		         copy;
		synchronized (bcsListeners) { copy = bcsListeners.toArray(); }

		for (int i = 0; i < copy.length; i++) {
			 org.omg.CORBA.Object c = (org.omg.CORBA.Object)copy[i];
			(BeanContextServicesListenerHelper.narrow(c)).serviceAvailable(bcssae);
		}
    }

    /**
     * Fire a BeanContextServiceEvent notifying of a service being revoked
     */

    protected final void fireServiceRevoked(BeanContextServiceRevokedEvent bcsre)
	throws IllegalArgumentException
	 {
		Object[]		         copy;
		synchronized (bcsListeners) { copy = bcsListeners.toArray(); }

		for (int i = 0; i < copy.length; i++) {
			org.omg.CORBA.Object c = (org.omg.CORBA.Object)copy[i];
			(BeanContextServicesListenerHelper.narrow(c)).serviceRevoked(bcsre);
		}
    }

    protected final void fireServiceRevoked(String serviceClass, 
		boolean revokeNow) throws IllegalArgumentException
  {
	Object[]		       copy;
	BeanContextServiceRevokedEventImpl bcsre = 
			new BeanContextServiceRevokedEventImpl(getBeanContextServicesPeer(),
										 serviceClass, revokeNow); 

	_BeanContextServiceRevokedEventImplBase_tie o_bcsre;
	 o_bcsre = new _BeanContextServiceRevokedEventImplBase_tie(bcsre);	  

	synchronized (bcsListeners) { copy = bcsListeners.toArray(); }

	for (int i = 0; i < copy.length; i++) {
		org.omg.CORBA.Object c = (org.omg.CORBA.Object)copy[i];
	    (BeanContextServicesListenerHelper.narrow(c)).serviceRevoked(o_bcsre);
	}
   }

    /**
     * called from BeanContextSupport writeObject before it serializes the
     * children ...
     *
     * This class will serialize any Serializable BeanContextServiceProviders
     * herein.
     *
     * subclasses may envelope this method to insert their own serialization
     * processing that has to occur prior to serialization of the children
     */

	/* TO DO LIST */
	/*
    protected synchronized void bcsPreSerializationHook(ObjectOutputStream oos) throws IOException {

	oos.writeInt(serializable);

	if (serializable <= 0) return;

	int count = 0;

	Iterator i = services.entrySet().iterator();

	while (i.hasNext() && count < serializable) {
	    Map.Entry 		entry = (Map.Entry)i.next();
	    BCSSServiceProvider bcsp  = null;

	     try {
		bcsp = (BCSSServiceProvider)entry.getValue();
	     } catch (ClassCastException cce) {
		continue;
	     }

	     if (bcsp.getServiceProvider() instanceof Serializable) {
		oos.writeObject(entry.getKey());
		oos.writeObject(bcsp);
		count++;
	     }
	}

	if (count != serializable) 
	    throw new IOException("wrote different number of service providers than expected");
    }
    */

    /**
     * called from BeanContextSupport readObject before it deserializes the
     * children ...
     *
     * This class will deserialize any Serializable BeanContextServiceProviders
     * serialized earlier thus making them available to the children when they
     * deserialized.
     *
     * subclasses may envelope this method to insert their own serialization
     * processing that has to occur prior to serialization of the children
     */

    /* TO DO LIST */
	/*
    protected synchronized void bcsPreDeserializationHook(ObjectInputStream ois) throws IOException, ClassNotFoundException {

	serializable = ois.readInt();

	int count = serializable;

	while (count > 0) {
	    services.put(ois.readObject(), ois.readObject());
	    count--;
	} 
    }
    */

    /**
     * serialize the instance
     */
	
	/* TO DO LIST */
	/*
    private synchronized void writeObject(ObjectOutputStream oos) throws IOException {
	oos.defaultWriteObject();

	serialize(oos, (Collection)bcsListeners);
    }
    */

    /**
     * deserialize the instance
     */

	/* TO DO LIST */
	/*
    private synchronized void readObject(ObjectInputStream ois) throws IOException, ClassNotFoundException {

	ois.defaultReadObject();

	deserialize(ois, (Collection)bcsListeners);
    }
    */

    
    /*
     * fields
     */

    /**
     * all accesses to the <code> protected transient HashMap services </code>
     * field should be synchronized on that object
     */

    protected transient myHashMap  		 services;

    protected transient int 	 		 serializable = 0;

    protected transient BCSSProxyServiceProvider proxy;

    protected transient ArrayList		 bcsListeners;
	org.omg.CORBA.ORB orb;
}



class myHashMap_with_is_equivalent extends HashMap{

    public myHashMap_with_is_equivalent(int initialCapacity)
	{
		super(initialCapacity);
	}

	public boolean containsKey(Object key) 
	{
	  org.omg.CORBA.Object elem;
	  org.omg.CORBA.Object okey = (org.omg.CORBA.Object)key;

	  Object[] olist = keySet().toArray();
	  for(int k=0;k<olist.length;k++){
		elem = (org.omg.CORBA.Object)olist[k];	
		if (elem._is_equivalent(okey)){
		    System.out.println("I found the object");
			return true;
		}
	  }
	  System.out.println("I  could NOT found the object");
	  return false;
	}
	
    public org.omg.CORBA.Object getKey(Object key) 
	{
	  org.omg.CORBA.Object elem;
	  org.omg.CORBA.Object okey = (org.omg.CORBA.Object)key;

	  Object[] olist = keySet().toArray();
	  for(int k=0;k<olist.length;k++){
		elem = (org.omg.CORBA.Object)olist[k];	
		if (elem._is_equivalent(okey)){
		    System.out.println("I found the object");
			return elem;
		}
	  }
	  System.out.println("I could NOT found the object");
	  return null;
	}


	public Object get(Object key)
	{
		org.omg.CORBA.Object rkey = getKey(key);
		if (rkey == null)
			return null;

		Object rval = super.get(rkey);
		return rval;
	}

	public Object remove(Object key)
	{
		org.omg.CORBA.Object rkey = getKey(key);
		if (rkey == null)
			return null;

		Object rval = super.remove(rkey);
		return rval;
	}



}
