package WebFlow;

import WebFlow.event.*;
import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import org.omg.CORBA.*;
import java.util.*;

/**
 * This is a utility class that can be used by beans that support bound
 * properties.  You can use an instance of this class as a member field
 * of your bean and delegate various work to it.
 */

public class PropertyChangeSupport 
{
  
    
    /**
     * @param sourceBean  The bean to be given as the source for any events.
     */

  public PropertyChangeSupport(org.omg.CORBA.Object sourceBean, 
			       BeanContextChildSupport BChild_)
  {
    
	if (sourceBean == null) {
	    throw new org.omg.CORBA.BAD_PARAM("SourceBean is Null in ProperChangeSupport");
	    //new NullPointerException();
	}
	source = sourceBean;

	Properties props = System.getProperties();
	props.put("org.omg.CORBA.ORBClass", "com.ooc.CORBA.ORB");
	props.put("org.omg.CORBA.ORBSingletonClass", "com.ooc.CORBA.ORBSingleton");
	System.setProperties(props);
	String []args = {};

	orb_ = org.omg.CORBA.ORB.init(args, props);

	//adaptor will be available through "getBeanContext" method
	adaptor_ = null;	
	BChild = BChild_;
   }

    
    public void insertContextInfo(Request request)
    {
	//System.out.println("PCS:insertContextInfo Entered");
	Context context = orb_.get_default_context();
	Any val = orb_.create_any();
	//add one method to BeanContextChild class , getIRid() and setIRid()
	String ID = BChild.getObjectID();
	val.insert_string(ID);
	context.set_one_value("ObjectID", val);

	//set Context of Request Object
	//the parameter of add method ("IR*") specifies which values of context
	//will be sent through. 
	//Here we say "sent all attributes starting with "IR"

	request.contexts().add("Object*");
	request.ctx(context);
	//System.out.println("PCS:insertContextInfo appended Context Info.:"+ID);
			  
    }


    /**
     * Add a PropertyChangeListener to the listener list.
     * The listener is registered for all properties.
     *
     * @param listener  The PropertyChangeListener to be added
     */

    public synchronized void addPropertyChangeListener(
				org.omg.CORBA.Object listener) 
   {
       //System.out.println("PropertyChangeSupport:addPropertyChangeListener:listener"+listener);
       if (listeners == null) {
	   listeners = new myHashMap(1); //java.util.Vector();
       }
       
       listeners.put(listener, listener);
       //if (listeners == null)
       //  System.out.println("PropertyChangeSupport :addPropertyChangeListener: What HAPPEDENED");
       //else
       //  System.out.println("PropertyChangeSupport :addPropertyChangeListener: Added"+
       //     " Listeners :sizE:"+listeners.size());	
    }


    /**
     * Remove a PropertyChangeListener from the listener list.
     * This removes a PropertyChangeListener that was registered
     * for all properties.
     *
     * @param listener  The PropertyChangeListener to be removed
     */

    public synchronized void removePropertyChangeListener(
				org.omg.CORBA.Object listener) {
	if (listeners == null) {
	    return;
	}

	listeners.remove(listener);
	//listeners.removeElement(listener);
	//System.out.println("PropertyChangeSupport :removePropertyChangeListener: Removed"+
	//	   " Listeners :sizE:"+listeners.size());
    }

    /**
     * Add a PropertyChangeListener for a specific property.  The listener
     * will be invoked only when a call on firePropertyChange names that
     * specific property.
     *
     * @param propertyName  The name of the property to listen on.
     * @param listener  The PropertyChangeListener to be added
     */

    public synchronized void addPropertyChangeListener(
				String propertyName,
				org.omg.CORBA.Object listener) 
    {
	System.out.println("PropertyChangeSupport:addPropertyChangeListener:"+
			   propertyName+" listener"+listener);
	if (children == null) {
	    children = new java.util.Hashtable();
	}
	PropertyChangeSupport child = (PropertyChangeSupport)children.get(propertyName);
	if (child == null) {
	    child = new PropertyChangeSupport(source, BChild);
	    children.put(propertyName, child);
	}
	child.addPropertyChangeListener(listener);

    }

   
    /**
     * Remove a PropertyChangeListener for a specific property.
     *
     * @param propertyName  The name of the property that was listened on.
     * @param listener  The PropertyChangeListener to be removed
     */

    public synchronized void removePropertyChangeListener(
				String propertyName,
				org.omg.CORBA.Object listener) {
	
	if (children == null) {
	    return;
	}
	PropertyChangeSupport child = (PropertyChangeSupport)children.get(propertyName);
	if (child == null) {
	    return;
	}
	//System.out.println("PropertyChangeSupport:REMOVEPropertyChangeListener:"+propertyName);
	child.removePropertyChangeListener(listener);
    }

    
    public void fireEvent_old(String eventMethod, org.omg.CORBA.Object evt)
    {
        //System.out.println("fireEvent is Called :"+eventMethod);
	adaptor_ = BChild.getBeanContext();

	//System.out.println("fireEvent: BeanContext:"+adaptor_);
	//You know we keeps the translation table in the adaptor. We don't need
	//to keep same table in event source side.
	if (adaptor_ != null){
	  invokeEventMethod(adaptor_, eventMethod, evt);
	  return;
	}
    }    

//Notice that the event name(id) specified in configuration file is same as eventMethod
//In this case, we will call method called "Event Id" on container object
public void fireEvent(String eventMethod, org.omg.CORBA.Object evt)
    {
        //System.out.println("fireEvent is Called ");
      
	myHashMap targets = null;
	PropertyChangeSupport child = null;
	synchronized (this) {
	    if (listeners != null) {
	        targets = (myHashMap) listeners.clone();
	    }
	    if (children != null && eventMethod != null) {
		child = (PropertyChangeSupport)children.get(eventMethod);
	    }
	}
       
	if (targets != null) {
	    for (java.util.Iterator elems = targets.values().iterator();elems.hasNext();) {
		org.omg.CORBA.Object target	= (org.omg.CORBA.Object)elems.next();
		invokeEventMethod(target, eventMethod, evt);
	    }
	}
	if (child != null) {
	    child.fireEvent(eventMethod, evt);
	}
  }

public void invokeEventMethod (org.omg.CORBA.Object target,
			       String eventMethod,
			       org.omg.CORBA.Object evt)
  {
    
    //System.out.println("Send Event from:"+BChild.getObjectID());

    org.omg.CORBA.Request request = target._request(eventMethod);
    request.add_in_arg().insert_Object(evt);

    //add my Object ID in Context
    insertContextInfo(request);

    request.invoke();
    java.lang.Exception e = request.env().exception();
    if(e != null)
      System.out.println("PropertychangeSupport:I got exception during :invokeEventMethod:"+ eventMethod+" :-: "+e);
    
  }

public void invokePropertyChange(org.omg.CORBA.Object target,
				 PropertyChangeEvent evt)
  {
    
    org.omg.CORBA.Request request = target._request("propertyChange");
    request.add_in_arg().insert_Object(evt);

    //add my Object ID in Context
    insertContextInfo(request);

    request.invoke();
    java.lang.Exception e = request.env().exception();
    if(e != null)
      System.out.println("I got exception during :invokePropertyChange"+e);
    
  }

/**
     * Report a bound property update to any registered listeners.
     * No event is fired if old and new are equal and non-null.
     *
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    public void firePropertyChange(String propertyName, 
				   org.omg.CORBA.Any oldValue, 
				   org.omg.CORBA.Any newValue) 
	{
         
	    //System.out.println("PropertyChangeSupport(Any):firePropertyChange :propName:"+propertyName);
		
		if (oldValue != null && newValue != null && oldValue.equal(newValue)) 
		{
			return;
		}
		PropertyChangeEvent evt = new PropertyChangeEventImpl(source,
					    propertyName, oldValue, newValue);
		sendEvent(propertyName, evt);
    }

    public void sendEvent_old(String propertyName, PropertyChangeEvent evt)
	{
		adaptor_ = BChild.getBeanContext();
		if (adaptor_ != null){
			invokePropertyChange(adaptor_, evt);
			return;
		}
	}

public void sendEvent(String propertyName, PropertyChangeEvent evt)
	{
	    //System.out.println("PropertyChangeSupport:SendEvent : ENTERED");

		myHashMap targets = null;
		PropertyChangeSupport child = null;
		synchronized (this) {
		    if (listeners != null) 
			targets = (myHashMap) listeners.clone();
		    if (children != null && propertyName != null) 
			child = (PropertyChangeSupport)children.get(propertyName);
		}

		
		if (targets != null) {
		    for (java.util.Iterator elems = targets.values().iterator();elems.hasNext();) {
			org.omg.CORBA.Object target	= (org.omg.CORBA.Object)elems.next();
			invokePropertyChange(target,evt);
				//target.propertyChange(evt);
		    }
		}

		if (child != null) {
		    //System.out.println("PropertyChangeSupport:SendEvent : child SIZE:");
		    child.firePropertyChange(evt);
		}	

	}

    //
    // IDL:/WebFlow/WebFlowContext/pullEvents:1.0
    //
    public org.omg.CORBA.Object[]
    invokePullEventsOfListeners(org.omg.CORBA.Object target)
    {
      if(target == null){
	System.out.println("Oops my Parent is Null any more");
	return null;
      }
        org.omg.CORBA.Request _ob_req = target._request("pullEvents");

        _ob_req.set_return_type(membersArrayHelper.type());

	//add my Object ID in Context
	insertContextInfo(_ob_req);

        _ob_req.invoke();

        java.lang.Exception _ob_ex = _ob_req.env().exception();
        if(_ob_ex != null)
            throw (org.omg.CORBA.SystemException)_ob_ex;

        org.omg.CORBA.Any _ob_ra = _ob_req.return_value();
        return membersArrayHelper.extract(_ob_ra);
    }

     //pull the stored events at the containers of target components
    public org.omg.CORBA.Object [] pull()
    {
	Vector vlist=new Vector();
	pullExtended(vlist);
	//System.out.println("PropertyChangeSuppor:pull:after pullExtended.");

       	int nelem = vlist.size();
	org.omg.CORBA.Object []elist=new org.omg.CORBA.Object[nelem];
	for(int k=0;k<nelem;k++)
	  elist[k] = (org.omg.CORBA.Object)vlist.elementAt(k);

	//System.out.println("PropertyChangeSuppor:pull:after elist.");
	return elist;
    }
    

    //pull the stored events at the containers of target components
    public void pullExtended(Vector vlist)
	{
	    String propertyName;
	    myHashMap targets = null;
	    Hashtable childs=null;
	    org.omg.CORBA.Object []elist;
	
	    //System.out.println("PropertyChangeSupport:pullExtended : ENTERED");
	    synchronized (this) {
		if (listeners != null) 
		    targets = (myHashMap) listeners.clone();
	    }

	    //pull events from listeners I was registered for all of their property changes
	    if (targets != null) {
		for (java.util.Iterator elems = targets.values().iterator();elems.hasNext();) {
		  //System.out.println(" PCS:pulling event");
		  org.omg.CORBA.Object target	= (org.omg.CORBA.Object)elems.next();
		  elist = invokePullEventsOfListeners(target);
		  // System.out.println("PropertyChangeSupport:pullExtended:elist:"+
		  //     elist.length);
		  vlist.addAll(Arrays.asList(elist));
		  // System.out.println("PropertyChangeSupport:pullExtended:vlist:"+
		  //	      vlist.size());
		}
	    }

	    //pull events from listeners I was registered for specific property changes
	    java.util.Iterator iter;
	    PropertyChangeSupport child = null;
	    
	    synchronized (this) {
		if (children != null)
		    childs = (Hashtable) children.clone();
	    }
	    if (childs == null){
	      //System.out.println(" PCS: childs null");
	      return ;
	    }
	    //else
	    // System.out.println(" PCS: childs NOOOOT null");
	    
	    for (java.util.Iterator elems = childs.values().iterator();elems.hasNext();) {
		child= (PropertyChangeSupport)elems.next();
		child.pullExtended(vlist);
	    }
	}


    public void listContents(String propertyName, Hashtable children)
    {
	System.out.println("LISTENER ENTERED for :"+propertyName);
	for(Enumeration keys = children.keys();keys.hasMoreElements();)
	    System.out.println("LISTENER for propertyname:"+(String)keys.nextElement());
	
    }

    /**
     * Report an int bound property update to any registered listeners.
     * No event is fired if old and new are equal and non-null.
     * <p>
     * This is merely a convenience wrapper around the more general
     * firePropertyChange method that takes Object values.
     *
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    public void firePropertyChange(String propertyName, 
					int oldValue, int newValue) {
	if (oldValue == newValue) {
	    return;
	}
	Any any_oldValue = orb_.create_any();
	any_oldValue.insert_long(oldValue);
	Any any_newValue = orb_.create_any();
	any_newValue.insert_long(newValue);

	firePropertyChange(propertyName, any_oldValue,any_newValue);
	//new Integer(oldValue), new Integer(newValue));
    }

    public void firePropertyChange(String propertyName, 
					String oldValue, String newValue) {
	if (oldValue == newValue) {
	    return;
	}
	Any any_oldValue = orb_.create_any();
	any_oldValue.insert_string(oldValue);
	Any any_newValue = orb_.create_any();
	any_newValue.insert_string(newValue);

	firePropertyChange(propertyName, any_oldValue,any_newValue);
	//new Integer(oldValue), new Integer(newValue));
    }
 
	 //fire object values	
    public void firePropertyChange(String propertyName, 
				   org.omg.CORBA.Object oldValue, 
				   org.omg.CORBA.Object newValue)
    {
      
	//System.out.println("1:firePropertyChange:Object : old&new:"
	//	 +oldValue+"&"+newValue);
    
		BeanContextChild old_bcc = BeanContextChildHelper.narrow(oldValue);
		BeanContextChild new_bcc = BeanContextChildHelper.narrow(newValue);
		String oldID = (old_bcc==null)?null:old_bcc.getObjectID();
		String newID = (new_bcc==null)?null:new_bcc.getObjectID();
		if ((oldID != null)&&(oldID.equals(newID))){
		    System.out.println("firePropertyChange:I am returning from firePropertyChange:is_quivalent");
		    return;
		}
		PropertyChangeEvent evt = new PropertyChangeEventImpl(source,
					    propertyName, oldValue, newValue);
		sendEvent(propertyName, evt);
    }


    /**
     * Report a boolean bound property update to any registered listeners.
     * No event is fired if old and new are equal and non-null.
     * <p>
     * This is merely a convenience wrapper around the more general
     * firePropertyChange method that takes Object values.
     *
     * @param propertyName  The programmatic name of the property
     *		that was changed.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    public void firePropertyChange(String propertyName, 
					boolean oldValue, boolean newValue) {
	if (oldValue == newValue) {
	    return;
	}
	Any any_oldValue = orb_.create_any();
	any_oldValue.insert_boolean(oldValue);
	Any any_newValue = orb_.create_any();
	any_newValue.insert_boolean(newValue);
	firePropertyChange(propertyName, any_oldValue, any_newValue);

	//firePropertyChange(propertyName, new Boolean(oldValue), new Boolean(newValue));
    }

    /**
     * Fire an existing PropertyChangeEvent to any registered listeners.
     * No event is fired if the given event's old and new values are
     * equal and non-null.
     * @param evt  The PropertyChangeEvent object.
     */
    public void firePropertyChange_old(PropertyChangeEvent evt)
	{
		org.omg.CORBA.Any oldValue = evt.getOldValue();
		org.omg.CORBA.Any newValue = evt.getNewValue();
		//System.out.println("2:firePropertyChange : old&new:"
		// +oldValue+"&"+newValue);
        String propertyName = evt.getPropertyName();
		if (oldValue != null && newValue != null && oldValue.equals(newValue)) {
			return;
		}
		sendEvent(propertyName, evt);
   }

    public void firePropertyChange(PropertyChangeEvent evt) {
	org.omg.CORBA.Any oldValue = evt.getOldValue();
	org.omg.CORBA.Any newValue = evt.getNewValue();


        String propertyName = evt.getPropertyName();
	if (oldValue != null && newValue != null && 
	    oldValue.equals(newValue)) {
	    return;
	}
		
	myHashMap targets = null;
	PropertyChangeSupport child = null;
	synchronized (this) {
	    if (listeners != null) {
	        targets = (myHashMap) listeners.clone();
	    }
	    if (children != null && propertyName != null) {
		child = (PropertyChangeSupport)children.get(propertyName);
	    }
	}
   
	
	if (targets != null) {
	    for (java.util.Iterator elems = targets.values().iterator();elems.hasNext();) {
		org.omg.CORBA.Object target	= (org.omg.CORBA.Object)elems.next();
		invokePropertyChange(target,evt);
		//target.propertyChange(evt);
	    }
	}
	if (child != null) {
	    child.firePropertyChange(evt);
	}
    }

    /**
     * Check if there are any listeners for a specific property.
     *
     * @param propertyName  the property name.
     * @return true if there are ore or more listeners for the given property
     */
    public synchronized boolean hasListeners(String propertyName) {
	if (listeners != null && !listeners.isEmpty()) {
	    // there is a generic listener
	    return true;
	}
	if (children != null) {
	    PropertyChangeSupport child = (PropertyChangeSupport)children.get(propertyName);
	    if (child != null) {
		// The child will always have a listeners Vector.		    
		return !child.listeners.isEmpty();
	    }
	}
	return false;
    }

    BeanContextChildSupport BChild;
    private org.omg.CORBA.ORB orb_;
	private org.omg.CORBA.Object adaptor_;
  
    // "listeners" lists all the generic listeners.
    //transient private java.util.Vector listeners;
    public myHashMap listeners;
    // "children" contains PropertyChangeSupports for individual properties
    public java.util.Hashtable children;
    private org.omg.CORBA.Object source;
    private int propertyChangeSupportSerializedDataVersion = 2;

    // Serialization version ID, so we're compatible with JDK 1.1
    static final long serialVersionUID = 6401253773779951803L;
}
