package WebFlow;

import WebFlow.event.*;

import java.io.Serializable;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import org.omg.CORBA.*;
import java.util.*;

/**
 * This is a utility class that can be used by beans that support constrained
 * properties.  You can use an instance of this class as a member field
 * of your bean and delegate various work to it.
 */

public class VetoableChangeSupport 
{
  
     /**
     * @param sourceBean  The bean to be given as the source for any events.
     */

    public VetoableChangeSupport(org.omg.CORBA.Object sourceBean,
								 BeanContextChildSupport BChild_)
								 throws NullPointerException
	{
	      Properties props = System.getProperties();
	  props.put("org.omg.CORBA.ORBClass", "com.ooc.CORBA.ORB");
	  props.put("org.omg.CORBA.ORBSingletonClass", "com.ooc.CORBA.ORBSingleton");
	  System.setProperties(props);
	  String []args = {};

		if (sourceBean == null) {
			throw new NullPointerException();
		}
		source = sourceBean;
		orb_ =  org.omg.CORBA.ORB.init(args, props);

		//adaptor will be available through "getBeanContext" method
		adaptor_ = null;
        BChild = BChild_;
	 }
	 
    /**
     * Add a VetoableListener to the listener list.
     * The listener is registered for all properties.
     *
     * @param listener  The VetoableChangeListener to be added
     */

    public synchronized void addVetoableChangeListener(
					org.omg.CORBA.Object listener) //throws NullPointerException 
	{
		if (listeners == null) {
		    listeners = new myHashMap(1);
		}
		listeners.put(listener,listener);
		//listeners.addElement(listener);
    }

    /**
     * Remove a VetoableChangeListener from the listener list.
     * This removes a PropertyChangeListener that was registered
     * for all properties.
     *
     * @param listener  The VetoableChangeListener to be removed
     */
    public synchronized void removeVetoableChangeListener(
					org.omg.CORBA.Object listener) 
	{
		if (listeners == null) {
			return;
		}
		listeners.remove(listener);
		//listeners.removeElement(listener);
		//System.out.println("VetoableChangeSupport:removePropertyChangeListener: Removed"
		//+   " Listeners :sizE:"+listeners.size());
    }


    /**
     * Add a VetoableChangeListener for a specific property.  The listener
     * will be invoked only when a call on fireVetoableChange names that
     * specific property.
     *
     * @param propertyName  The name of the property to listen on.
     * @param listener  The VetoableChangeListener to be added
     */

    public synchronized void addVetoableChangeListener(
				String propertyName,
				org.omg.CORBA.Object listener) throws NullPointerException
	{
		if (children == null) {
			children = new java.util.Hashtable();
		}
		VetoableChangeSupport child = (VetoableChangeSupport)children.get(propertyName);
		if (child == null) {
			child = new VetoableChangeSupport(source, BChild);
			children.put(propertyName, child);
		}
		child.addVetoableChangeListener(listener);
    }

   
    /**
     * Remove a VetoableChangeListener for a specific property.
     *
     * @param propertyName  The name of the property that was listened on.
     * @param listener  The VetoableChangeListener to be removed
     */

    public synchronized void removeVetoableChangeListener(
				String propertyName,
				org.omg.CORBA.Object listener) {
	if (children == null) {
	    return;
	}
	VetoableChangeSupport child = (VetoableChangeSupport)children.get(propertyName);
	if (child == null) {
	    return;
	}
	child.removeVetoableChangeListener(listener);
    }


    /**
     * Report a vetoable property update to any registered listeners.  If
     * anyone vetos the change, then fire a new event reverting everyone to 
     * the old value and then rethrow the PropertyVetoException.
     * <p>
     * No event is fired if old and new are equal and non-null.
     *
     * @param propertyName  The programmatic name of the property
     *		that is about to change..
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     * @exception PropertyVetoException if the recipient wishes the property
     *              change to be rolled back.
     */
    public void fireVetoableChange(String propertyName, 
				   org.omg.CORBA.Any oldValue,
				   org.omg.CORBA.Any newValue)
      throws PropertyVetoException 
	{
		
       	PropertyChangeEvent evt = new PropertyChangeEventImpl(source, propertyName,
							    oldValue, newValue);
	    fireVetoableChange(evt);
    }

    public void fireVetoableChange(String propertyName, 
					String oldValue, String  newValue)
					throws PropertyVetoException {
	if (oldValue == newValue) {
	    return;
	}
	Any any_oldValue = orb_.create_any();
	any_oldValue.insert_string(oldValue);
	Any any_newValue = orb_.create_any();
	any_newValue.insert_string(newValue);
	fireVetoableChange(propertyName, any_oldValue,any_newValue);

	//fireVetoableChange(propertyName, new Integer(oldValue), new Integer(newValue));
    }


    /**
     * Report a int vetoable property update to any registered listeners.
     * No event is fired if old and new are equal and non-null.
     * <p>
     * This is merely a convenience wrapper around the more general
     * fireVetoableChange method that takes Object values.
     *
     * @param propertyName  The programmatic name of the property
     *		that is about to change.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    public void fireVetoableChange(String propertyName, 
					int oldValue, int newValue)
					throws PropertyVetoException {
	if (oldValue == newValue) {
	    return;
	}
	Any any_oldValue = orb_.create_any();
	any_oldValue.insert_long(oldValue);
	Any any_newValue = orb_.create_any();
	any_newValue.insert_long(newValue);
	fireVetoableChange(propertyName, any_oldValue,any_newValue);

	//fireVetoableChange(propertyName, new Integer(oldValue), new Integer(newValue));
    }

    /**
     * Report a boolean vetoable property update to any registered listeners.
     * No event is fired if old and new are equal and non-null.
     * <p>
     * This is merely a convenience wrapper around the more general
     * fireVetoableChange method that takes Object values.
     *
     * @param propertyName  The programmatic name of the property
     *		that is about to change.
     * @param oldValue  The old value of the property.
     * @param newValue  The new value of the property.
     */
    public void fireVetoableChange(String propertyName, 
					boolean oldValue, boolean newValue) 
					throws PropertyVetoException {
	if (oldValue == newValue) {
	    return;
	}
	Any any_oldValue = orb_.create_any();
	any_oldValue.insert_boolean(oldValue);
	Any any_newValue = orb_.create_any();
	any_newValue.insert_boolean(newValue);
	fireVetoableChange(propertyName, any_oldValue, any_newValue);


	//fireVetoableChange(propertyName, new Boolean(oldValue), new Boolean(newValue));
    }


public void insertContextInfo(Request request)
    {
	//System.out.println("PCS:insertContextInfo Entered");
	Context context = orb_.get_default_context();
	Any val = orb_.create_any();
	//add one method to BeanContextChild class , getIRid() and setIRid()
	String ID = BChild.getObjectID();
	val.insert_string(ID);
	context.set_one_value("ObjectID", val);

	//set Context of Request Object
	//the parameter of add method ("IR*") specifies which values of context
	//will be sent through. 
	//Here we say "sent all attributes starting with "IR"

	request.contexts().add("Object*");
	request.ctx(context);
	//System.out.println("PCS:insertContextInfo appended Context Info.:"+ID);
			  
    }

   public void invokeVetoablePropertyChange(org.omg.CORBA.Object target,
				 PropertyChangeEvent evt)
                           throws PropertyVetoException
  {
    
    org.omg.CORBA.Request request = target._request("vetoablePropertyChange");
    request.add_in_arg().insert_Object(evt);
    request.exceptions().add(WebFlow.event.PropertyVetoExceptionHelper.type());
   
    //add my Object ID in Context
    insertContextInfo(request);
  
    request.invoke();
    java.lang.Exception _ob_ex = request.env().exception();
    if(_ob_ex != null)
      {
	org.omg.CORBA.UnknownUserException _ob_uex;
	try
	  {
	    _ob_uex = (org.omg.CORBA.UnknownUserException)_ob_ex;
	  }
	catch(ClassCastException _ob_dummy)
	  {
	    throw (org.omg.CORBA.SystemException)_ob_ex;
	  }
	System.out.println("VCSupp:I got exception :invokeVetoable"+
             "PropertyChange: "+_ob_ex);
	try
	  {
	    throw WebFlow.event.PropertyVetoExceptionHelper.extract(_ob_uex.except);
	  }
	catch(org.omg.CORBA.BAD_OPERATION _ob_dummy)
	  {
	  }

	throw new org.omg.CORBA.UNKNOWN();
      }
  }

    //send object values
     public void fireVetoableChange(String propertyName, 
					org.omg.CORBA.Object oldValue,
					org.omg.CORBA.Object newValue) 
					throws PropertyVetoException 
	{

	    //System.out.println("fireVetoableChange:entered:old:"+oldValue+
	    //			" new:"+newValue);
	    BeanContextChild old_bcc = BeanContextChildHelper.narrow(oldValue);
	    BeanContextChild new_bcc = BeanContextChildHelper.narrow(newValue);
	    String oldID = (old_bcc==null)?null:old_bcc.getObjectID();
	    String newID = (new_bcc==null)?null:new_bcc.getObjectID();
	    if ((oldID != null)&&(oldID.equals(newID))){
		System.out.println("I am returning from firePropertyChange:is_quivalent");
		return;
	    }

	    PropertyChangeEvent evt = new PropertyChangeEventImpl(source, propertyName,
							    oldValue, newValue);

	    //System.out.println("fireVetoableChange:before sendEvent");
	    sendEvent(propertyName,evt);
	    //System.out.println("fireVetoableChange:after sendEvent");
    }

    /**
     * Fire a vetoable property update to any registered listeners.  If
     * anyone vetos the change, then fire a new event reverting everyone to 
     * the old value and then rethrow the PropertyVetoException.
     * <p>
     * No event is fired if old and new are equal and non-null.
     *
     * @param evt  The PropertyChangeEvent to be fired.
     * @exception PropertyVetoException if the recipient wishes the property
     *              change to be rolled back.
     */
    public void fireVetoableChange(PropertyChangeEvent evt)
					throws PropertyVetoException 
	{

       //System.out.println("fireVetoableChange CALLED..");

		org.omg.CORBA.Any oldValue = evt.getOldValue();
		org.omg.CORBA.Any newValue = evt.getNewValue();
		String propertyName = evt.getPropertyName();
		if (oldValue != null && newValue != null && oldValue.equal(newValue)) {
			return;
		}
		//send event
		sendEvent(propertyName, evt);
    }


    public void sendEvent_old(String propertyName, PropertyChangeEvent evt)
	  throws PropertyVetoException
	{

	    adaptor_ = BChild.getBeanContext();
		org.omg.CORBA.Any oldValue = evt.getOldValue();
	    org.omg.CORBA.Any newValue = evt.getNewValue();

		if (listeners != null) {
		  try {
			  if (adaptor_ != null)
				invokeVetoablePropertyChange(adaptor_, evt);
		  } catch (PropertyVetoException veto) {
	        // Create an event to revert everyone to the old value.
			//newly created event is same as argument of this method, isn't it?
	        System.out.println("fireVetoableChange got Exception..");
	        evt = new PropertyChangeEventImpl(source, propertyName, newValue, oldValue);
	        if (adaptor_ != null)
				invokeVetoablePropertyChange(adaptor_, evt);
	        // And now rethrow the PropertyVetoException.
	        throw veto;
		  } //catch
	    } //if (listener ..)
   }


	public void sendEvent(String propertyName, PropertyChangeEvent evt)
	  throws PropertyVetoException
	{
	    org.omg.CORBA.Any oldValue = evt.getOldValue();
	    org.omg.CORBA.Any newValue = evt.getNewValue();

	    myHashMap targets = null;
	    VetoableChangeSupport child = null;
	    synchronized (this) {
		if (listeners != null) {
		    targets = (myHashMap) listeners.clone();
		}
		if (children != null && propertyName != null) {
		    //listContents(propertyName, children);
		    child = (VetoableChangeSupport)children.get(propertyName);
		    //System.out.println("VEtoabel child object:"+child);
		}
	    }

	    if (listeners != null) {
		try {
		    for (java.util.Iterator elems = targets.values().iterator();elems.hasNext();) {
			org.omg.CORBA.Object target	= (org.omg.CORBA.Object)elems.next();
			invokeVetoablePropertyChange(target,evt);
			//target.vetoableChange(evt);
		    } //for
		} catch (PropertyVetoException veto) {
		    // Create an event to revert everyone to the old value.
		    //newly created event is same as argument of this method, isn't it?
		    System.out.println("fireVetoableChange got Exception..");
		    evt = new PropertyChangeEventImpl(source, propertyName, newValue, oldValue);
		    for (java.util.Iterator elems = targets.values().iterator();elems.hasNext();) {
			try{
			    org.omg.CORBA.Object target	= (org.omg.CORBA.Object)elems.next();
			    //System.out.println("VCSupport Revert back :"+i);
			    invokeVetoablePropertyChange(target,evt);
			        //target.vetoableChange(evt);
			} catch (PropertyVetoException ex) {
			    // We just ignore exceptions that occur during reversions.
			} //catch
		    }//for 
		    // And now rethrow the PropertyVetoException.
		    throw veto;
		} //catch
	    }//if (listener ..)

    if (child != null) {
	child.fireVetoableChange(evt);
    }
}

   public void listContents(String propertyName, Hashtable children)
    {
	System.out.println("VETOABLE LISTENER ENTERED for :"+propertyName);
	for(Enumeration keys = children.keys();keys.hasMoreElements();)
	    System.out.println("VETOABLE LISTENER for propertyname:"+(String)keys.nextElement());
	
    }

    /**
     * Check if there are any listeners for a specific property.
     *
     * @param propertyName  the property name.
     * @return true if there are one or more listeners for the given property
     */
    public synchronized boolean hasListeners(String propertyName) {
	if (listeners != null && !listeners.isEmpty()) {
	    // there is a generic listener
	    return true;
	}
	if (children != null) {
	    VetoableChangeSupport child = (VetoableChangeSupport)children.get(propertyName);
	    if (child != null) {
		// The child will always have a listeners Vector.		    
		return !child.listeners.isEmpty();
	    }
	}
	return false;
    }

	BeanContextChildSupport BChild;
    private org.omg.CORBA.ORB orb_;
	private org.omg.CORBA.Object adaptor_;

    // "listeners" lists all the generic listeners.
    transient private myHashMap listeners;
    // "children" contains VetoableChangeSupports for individual properties
    private java.util.Hashtable children;
    private org.omg.CORBA.Object source;
    private int vetoableChangeSupportSerializedDataVersion = 2;

    // Serialization version ID, so we're compatible with JDK 1.1
    static final long serialVersionUID = -5090210921595982017L;
}
