
package WebFlow;
import WebFlow.event.*;

/**
 * This class manages hookups between properties, so that a
 * bound property change on object X turns into a property
 * set on a related property on object Y.
 * <P>
 * We do this by associating a PropertyHookup adaptor with each
 * source object that we are interested in.  As part of the adaptor
 * we keep track of which target   setter methods to call when a given
 * property changes.
 */

import java.lang.reflect.*;
import java.io.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.*;

import org.omg.CORBA.*;


public class WebFlowContextImpl extends org.omg.CORBA.DynamicImplementation
						  													
{
  
  public _WebFlowContextImplBase_tie WebFlowContextPeer_tie;
  Vector proxyList;
  String WFSname_;
  printProcessOut ppo ;
  private boolean init=false;
  public moduleFactory  factory, RedMFImpl, BlackMFImpl;
  private moduleFactory  MFImpl;
  private configReader configSettings_;
  private org.omg.CORBA.ORB orb_;
  BOA boa_;
  // Table that maps from property names to a vector of HookupTargets
  Hashtable targetsForHookupSource;
  WebFlowContextOps wfo; 
  int myColor;
  Process RedFactory;
  public Process repository;
  public boolean non_existent ;

    /**
     * Constructor for a new property hookup adaptor.
     */
	//This constructor should be used to start WebFlow server
	//This is outermost context
    public WebFlowContextImpl(String configFile, ORB orb, BOA boa,Process repository)
       throws NullPointerException, PropertyVetoException
    {
        this.repository = repository;
	orb_  = orb;
        boa_  = boa;
	//This class will make local call to configReader class
	//Other objects will make remote call
	configSettings_ = new configReaderImpl(this, orb_, configFile);
	orb_.connect(configSettings_);
	
	//if one server is running, kill myself now
	configReaderImpl conf_impl = (configReaderImpl)configSettings_;
	non_existent = conf_impl.non_existent;
	if (!non_existent){
	  repository.destroy();
	  return;
	}
	
	//store the webFlow server name
	WFSname_ = configSettings_.getWFSName();

	// cretae a instance of _WebFlowContextImplBase_tie class to delegate 
	// invoke method
	wfo = new WebFlowContextOps(this,configSettings_, orb_, WFSname_);
	WebFlowContextPeer_tie = new _WebFlowContextImplBase_tie(wfo);
	orb_.connect(WebFlowContextPeer_tie);
	orb_.connect(this);

	new registerSlave();

	//Set my ObjectID
	wfo.setObjectID(WFSname_); //"/"+WFSname_);
	 
	//start module factory and wait it calls "start" method
	//so that "init" will be true;
	startModuleFactory();

	//Should we allow only master to has proxies or also slaves have proxies
	if (conf_impl.isMaster())
	  wfo.insertProxy(WFSname_);

	System.out.println("SERVER IS READY FOR ACCEPTING USER COMMANDS.....");
    }

   
    //This constructor should be used to insert subcontexts
    public WebFlowContextImpl(configReader configSettings, 
		ORB orb, String WFSName, moduleFactory factory)
	  throws NullPointerException, PropertyVetoException
    {
	System.out.println(WFSName+" build constructor of WebFlowContextImpl");
	//store the webFlow server name
	WFSname_ = WFSName;
    
	orb_  = orb;
	configSettings_ = configSettings;

	// cretae a instance of _WebFlowContextImplBase_tie class to delegate 
	// invoke method
	wfo = new WebFlowContextOps(this,configSettings_, orb_, WFSname_,factory);

	WebFlowContextPeer_tie = new _WebFlowContextImplBase_tie(wfo);
	orb_.connect(WebFlowContextPeer_tie);

	//set mycolor
	this.myColor = factory.getColor();
	this.factory  = factory;

	System.out.println("SERVER IS READY FOR ACCEPTING USER COMMANDS .....");
    }

  class registerSlave implements Runnable{

    public registerSlave()
    {
      Thread thr = new Thread(this);
      thr.start();
    }

    public void run()
  {
       System.out.println("Slave WebFlow :register isertme:enter");
       configReaderImpl conf_impl = (configReaderImpl)configSettings_;
        System.out.println("Slave WebFlow :register isertme:enter:"+conf_impl.isMaster());
	System.out.println("Slave WebFlow :register isertme:enter:"+conf_impl.isItMaster);
       if (conf_impl.isMaster())
	 return;

	//register myselt to master as a slave
	System.out.println("Slave WebFlow :register himself as slave to"+
			   conf_impl.masterServerIOR);
	WebFlowContext master = WebFlowContextHelper.narrow(
				    orb_.string_to_object(conf_impl.masterServerIOR));
	//master =  WebFlowContextHelper.narrow(master.getNamedContext(parentName));

	System.out.println("Slave WebFlow :after read master IOR:"+master);
	String objectID = conf_impl.parentName+"/"+conf_impl.ServerName;
	wfo.setObjectID(objectID);
	master.insertRemoteServer(conf_impl.ServerName,objectID,conf_impl.server_);
	System.out.println("Slave WebFlow :sent addWFServer request");
	//set myself as A WebFlow Server
	wfo.setAsServer();

  }
  }
	public String[]
    _ids()
    {
        return _WebFlowContextImplBase._ob_ids_;
    }
   
   public WebFlowContextOps getImpl()
  {
    return wfo;
  }

    public void invoke(org.omg.CORBA.ServerRequest request) 
    {
      String opname = request.op_name();
     
      //System.out.println("WebFlowContextImpl :"+opname+" is Called.:context:"+WFSname_);

	  //do propertyEvent forwarding and other administration jobs,
      //like attach property and event binding
      try{
	  wfo.setServerRequest(request);
	  WebFlowContextPeer_tie.invoke(request); //super.invoke(request);
	  return;
      }catch(org.omg.CORBA.BAD_OPERATION e){
	//System.out.println("WebFlowContextImpl:request: "+opname+  " came into WebFlowContextImpl.:context:"+WFSname_);
      }
      //System.out.println("Now I will make DII for request:"+opname+"-");
	  
      //forward Incoming Event
      wfo.forwardEvent(request);
    }
  
   public void checkSource(org.omg.CORBA.Object o)
    {
	//WebFlow.test.hello h = WebFlow.test.helloHelper.narrow(o);
	//h.sayHello();
      
    }
  

public void compareSourceTarget(org.omg.CORBA.Object target)
  {
    BeanContextChild ta = BeanContextChildHelper.narrow(target);
    WebFlowContext wta = WebFlowContextHelper.narrow(ta.getBeanContext());

   System.out.println("SOURCE COLOR:"+myColor+
		      "TARGET COLOR:"+wta.getMyColor());
       
  }
  
   public org.omg.CORBA.Object getSource(org.omg.CORBA.Object src)
  {
    org.omg.CORBA.ORB orb = orb_;//org.omg.CORBA.ORB.init();
    org.omg.CORBA.Request request = src._request("getSource");
    request.set_return_type(orb.create_interface_tc
		      ("IDL:omg.org/CORBA/Object:1.0","Object"));
    request.invoke();
    java.lang.Exception e = request.env().exception();
    if(e != null)
      System.out.println("I got exception : in getSource req. in Binder"+e);
    org.omg.CORBA.Object esrc = request.result().value().extract_Object();
    return esrc;
       
  }

   public void invokeSetterMethod(org.omg.CORBA.Object target,
				 String setter,
				 org.omg.CORBA.Object evt)
  {
    
    org.omg.CORBA.Request request = target._request(setter);
    request.add_in_arg().insert_Object(evt);
    request.invoke();
    java.lang.Exception e = request.env().exception();
    if(e != null)
      System.out.println("PROP:I got exception during :invokeSetterMethod: "+
			 setter+" :"+e);
  }

public void startModuleFactory()
{

  System.out.println(WFSname_+" is about starting moduleFactory");

  org.omg.CORBA.Object IRobj=null;
  String myIOR = orb_.object_to_string(this);
   
  try{
    IRobj = orb_.resolve_initial_references("InterfaceRepository");
  }catch(org.omg.CORBA.ORBPackage.InvalidName ex){
    System.out.println("Error in getting IR IOR in WFS:"+ex);
  }

  System.out.println(WFSname_+":moduleFactory gets ior :"+IRobj);

  String IRstr = orb_.object_to_string(IRobj);
   System.out.println(WFSname_+":aftre Irstr");
  String ConfigStr = orb_.object_to_string(configSettings_);

  try{
    //start RED Process
    String d = " -Dorg.omg.CORBA.ORBSingletonClass=com.ooc.CORBA.ORBSingleton -Dorg.omg.CORBA.ORBClass=com.ooc.CORBA.ORB ";
    String command = "java"+d+" WebFlow.FacServer " + myIOR+ " "+ ConfigStr+ 
      " "+String.valueOf(FactoryImpl.RED)+    " -ORBrepository "+IRstr;
    Process p = Runtime.getRuntime().exec(command);
    RedFactory = p;
    ppo = new printProcessOut(p);
    String s = ppo.getIOROfProcess();
    RedMFImpl = moduleFactoryHelper.narrow(orb_.string_to_object(s));
    ppo.startProcess();
    System.out.println(WFSname_+":started RED Factory");


    //Black factory will run in main process of WebFlow Server
    //BlackMFImpl = new FactoryImpl(myIOR,ConfigStr,FactoryImpl.BLACK);
    BlackMFImpl = new  FactoryImpl(myIOR,ConfigStr,FactoryImpl.BLACK,
				   orb_, boa_);

    System.out.println(WFSname_+":started BLACK Factory");
   
    //Inform each one of pair from another
    RedMFImpl.setOtherMFImpl(RedMFImpl,FactoryImpl.RED);
    RedMFImpl.setOtherMFImpl(BlackMFImpl, FactoryImpl.BLACK);
        
    BlackMFImpl.setOtherMFImpl(RedMFImpl,FactoryImpl.RED);
    BlackMFImpl.setOtherMFImpl(BlackMFImpl, FactoryImpl.BLACK);
    
    //inform worker object of I chosed initially Red color 
    //and Red module Factory
    myColor = FactoryImpl.BLACK;
    MFImpl = BlackMFImpl;
    wfo.setClorAndMFI(myColor, MFImpl);
	

  }catch(SecurityException e){
	System.out.println("WebFlowContextImpl:Error(Security):"+e);

  }catch (Exception ex){
    System.out.println("WebFlowContextImpl:Error in starting ModuleFactory :"+ex);
  }
//System.out.println("WebFlowContextImpl:"+WFSname_+ ": WFS startted module factory.");
 }


  

}//End of WebFlowContextImpl Class




