package WebFlow.RemoteFile;

import org.omg.CORBA.*;
import java.io.*;
import WebFlow.RemoteFile.RemoteFilePackage.*;

/**
* FileTransfer is used to send and receive files to and from remote host
* It uses RemoteFile CORBA object to work with the remote file
**/
public class FileTransfer
{
  private ORB orb;
  private RemoteFile file;
  private short block_size = 8912;
  private FTcallback callback;
  
  /**
  * FileTransfer constructor accepts ORB and RemoteFile IOR
  **/
  public FileTransfer(ORB anOrb, String obj_ior)
  {
    orb = anOrb;
    org.omg.CORBA.Object obj = orb.string_to_object(obj_ior);
    file = RemoteFileHelper.narrow(obj);
  }

  public FileTransfer(RemoteFile rf)
  {
    file = rf;
  }  
 
  /**
  * FileTransfer constructor accepts ORB and RemoteFile IOR
  **/
  public FileTransfer(ORB anOrb, String obj_ior, FTcallback aCallback) 
  {
    callback = aCallback;
    orb = anOrb;
    org.omg.CORBA.Object obj = orb.string_to_object(obj_ior);
    file = RemoteFileHelper.narrow(obj);
  }
  
  /**
  * FileSend sends local file to the remote host
  * @param source name of the local file to be sent
  * @param dest name of the remote file
  **/
  public void FileSend(String source, String dest) throws FileNotFoundException, FileException, IOException
  {
    //open local file
    File sfile = new File(source);
    FileInputStream source_file = new FileInputStream(sfile);
    long file_size = sfile.length();
    long pos = 0;
    byte buffer[] = new byte[block_size]; //create temporary buffer
    try {
      file.open(dest, "rw"); //open remote file to be written
      while(true) {
        int icount = source_file.read(buffer); //read data from the local file
        if(icount == -1)
          throw new EOFException();
        byte arr[];
        if(block_size == icount)
          arr = buffer;
        else {
          arr = new byte[icount]; //create buffer to be send to the remote host
          for(int i = 0; i < icount; i++) //copy temporary buffer
            arr[i] = buffer[i];
        }
        file.putBlock(arr); //write data to the remote file
        pos += icount;
        if(callback != null)
          callback.transfered((int)(pos * 100/file_size));
      }
    } 
    catch(EOFException ex) {
      file.close();            
      source_file.close();
        if(callback != null)
          callback.transfered(100);      
    }
    catch(IOException ex) {
      file.close();
      throw ex;
    }
  }
  
  /**
  * FileReceive copies file from the remote host to the local file
  * @param dest name of the local file to copy to
  * @param source name of the remote file to be copied
  **/
    public void FileReceive(String dest, String source) throws FileException, IOException
    {
	//open local file for writing
	java.io.FileOutputStream dest_file = new FileOutputStream(dest);
	//create buffer to hold read data
	byte arr[] = null;
	try {
	    file.open(source, "r"); //open remote file for reading
	    long file_size = file.getFileSize();
	    long pos = 0;
	    while(true) {
		arr = file.getBlock();
		if(arr.length > 0) {
		    dest_file.write(arr); //write block to the local file
		    pos += arr.length;
		    if(callback != null)
			callback.transfered((int)(pos * 100/file_size));
		}
		else throw new EOFFileException();
	    }
	}
	catch(EOFFileException ex) {
	    dest_file.flush(); //flush buffers
	    dest_file.close();
	    if(callback != null)
		callback.transfered(100);
	    file.close();
	    }
    }


    /*public void FileReceive(String dest, String source) throws FileException, IOException
  {  
    java.io.FileOutputStream dest_file = null;
    try {
      //open local file for writing
      dest_file = new FileOutputStream(dest);    
      //create buffer to hold read data
      octet_sequenceHolder arr = new octet_sequenceHolder();
      file.open(source, "r"); //open remote file for reading
      long file_size = file.getFileSize();
      long pos = 0;
      while(true) {
        file.getBlock(arr); //read block from the remote file
        dest_file.write(arr.value); //write block to the local file
        pos += arr.value.length;
        if(callback != null)
          callback.transfered((int)(pos * 100/file_size));
      }
    } 
    catch(EOFFileException ex) {
      dest_file.flush(); //flush buffers
      dest_file.close();
      if(callback != null)
        callback.transfered(100);
      file.close();            
    }
  }
    */
  
  /**
  * SetBlockSize sets size of the block of data to be sent for each transfer
  * @param aSize new size of the block
  **/
  public void SetBlockSize(short aSize)
  {
    block_size = aSize;
    file.setBlockSize(aSize);
  }
}
