package WebFlow.RemoteFile;

import java.io.*;
import WebFlow.RemoteFile.RemoteFilePackage.*;
import WebFlow.event.*;
import WebFlow.*;


/**
*
* RemoteFileImpl implements RemoteFile interface that allows user to
* manipulate with the remote file
*
**/
public class RemoteFileImp extends WebFlow.BeanContextChildSupport
                    implements RemoteFileOperations
{
  private RandomAccessFile file = null;
  private short block_size = 1024;
  org.omg.CORBA.Object peer;

  public RemoteFileImp(org.omg.CORBA.Object peer, String msg)
                                    throws WebFlow.NullPointerException
  {
    super(peer);
    this.peer = peer;
  }

  
  /**
  * open file specified by user.
  * @param file_name name of the file to open
  * @parem options type of access: "r" - read only, "rw" - read and write
  **/
  public void open(String file_name, String options) throws FileException
  {
    try {
      // try to open file with the specified options, raise FileException
      // if fails
      file = new RandomAccessFile(file_name,  options);
    } catch(java.io.IOException ex) { 
      throw new FileException("File: " + ex.getMessage());
      };
  }
  
  /**
  * close previously opened file.
  **/
  public void close()
  {
    try {
      file.close();
    } catch(java.io.IOException ex) { 
    }
    file = null;
  }
    
  /**
  * gets one byte from the file and returns it to the user
  * @param c holds read byte
  **/
  public void get(org.omg.CORBA.ByteHolder c) throws FileException, EOFFileException
  {
    //raise FileException if file is not opened
    if(file == null) throw new FileException("File: file is not opened");
    try {
      c.value = file.readByte(); //read one byte from the file
    }
    catch(java.io.EOFException ex) {
      //raise EOFFileException to notify user that end of file is reached
      throw new EOFFileException(); 
    }
    catch(java.io.IOException ex) { 
      throw new FileException("File: " + ex.getMessage());
    }
  }
  
  /**
  * reads block of bytes from the file and returns it to the user
  * @param arr sequence of read bytes. use arr.value.length to determine
  * how many bytes were read.
  **/
    /*
  public void getBlock(octet_sequenceHolder arr) throws FileException, EOFFileException
  {
      System.out.println("REMOTE FILE GETBLOCK BEGIN");
    if(file == null) throw new FileException("File: file is not opened");    
    try {
     //create temporary buffer to read bytes to
     byte buffer[] = new byte[block_size];      
     int icount = file.read(buffer);//read data
     if(icount == -1) { //if end of file is reached
      arr.value = new byte[1]; //dummy array is created to avoid marshaling error
      throw new EOFException();
     }
     if(icount == block_size) //if we have filled all the buffer
      arr.value = buffer;
     else { //if not all the buffer is used
      arr.value = new byte[icount]; //create array to return to user
      for(int i = 0; i < icount; i++) //copy temp array to the user array
       arr.value[i] = buffer[i];
     }
    }
    catch(java.io.EOFException ex) {
      //throw OEFFileException to indicate user that eof is reached
	System.out.println("REMOTE FILE THROWN EOF EXCEPTION");
      throw new EOFFileException();
    }
    catch(java.io.IOException ex) {
      throw new FileException("File: " + ex.getMessage());
    }
  }
    */

    public byte[] getBlock() throws FileException, EOFFileException
    {
	byte arr[] = null;
	if(file == null) throw new FileException("File: file is not opened");    
	try {
	    //create temporary buffer to read bytes to
	    byte buffer[] = new byte[block_size];      
	    int icount = file.read(buffer);//read data
	    if(icount == -1) { //if end of file is reached
		arr = new byte[0]; //dummy array is created to avoid marshaling error
		throw new EOFException();
	    }
	    if(icount == block_size) //if we have filled all the buffer
		arr = buffer;
	    else { //if not all the buffer is used
		arr = new byte[icount]; //create array to return to user
		for(int i = 0; i < icount; i++) //copy temp array to the user array
		    arr[i] = buffer[i];
	    }
	}
	catch(java.io.EOFException ex) {
	    //throw OEFFileException to indicate user that eof is reached
	    return arr;
	    // throw new EOFFileException();
	}
	catch(java.io.IOException ex) {
	    return arr;
	    //throw new FileException("File: " + ex.getMessage());
	}
	return arr;
    }

    /**
     * write one byte to the file
     * @param c value to be written to the file
     **/
    public void put(byte c) throws FileException
    {
	if(file == null) throw new FileException("File: file is not opened");
	try {
	    file.writeByte(c); //write byte to the file
	} catch(java.io.IOException ex) {
	    throw new FileException("File: " + ex.getMessage());
	}
    }
    
    /**
     * write block of bytes to the file
     * @param arr array of bytes to be written to the file.
     **/
    public void putBlock(byte arr[]) throws FileException
    {
	if(file == null) throw new FileException("File: file is not opened");
	try {
	    file.write(arr); //write array of bytes to the file
	}
	catch(java.io.IOException ex) {
	    throw new FileException("File: " + ex.getMessage());
	}
    }
  
    /**
     * set block size for reading file
     * @param aSize size of the block of octets to be read at one time
     **/  
    public void setBlockSize(short aSize)
    {
	block_size = aSize;
    }
  
    /**
     * Private function that is used to display status messages to stdout
     **/
    private void status(String msg)
    {
	System.out.println(msg);
    }
    
    // gets the size of the file
    public int getFileSize() throws FileException
    {
	//raise FileException if file is not opened
	if(file == null) throw new FileException("File: file is not opened");
	
	try {
	    return (int)file.length();
	}
	catch(java.io.IOException ex) {
	    throw new FileException("File: " + ex.getMessage());
	}
    }

    // deletes file or directory specified by file_name
    // directory must be empty to be deleted
    // use rmdir to delete not empty directory
    public boolean delete(String file_name)
    {
	try {
	    File file = new File(file_name);
	    file.delete();
	    return true;
	} catch (Exception ex) {
	    error(ex.getMessage());
	    return false;
	}
    }

    // creates directory with the specified name
    public boolean mkdir(String dir_name)
    {
	try {
	    File file = new File(dir_name);
            if(file.exists()) {
                return false;
	    }
            else {
	        file.mkdir();
                return true;
	    }
	} catch (Exception ex) {
	    error(ex.getMessage());
	    return false;
	}
    }

    // deletes directory with the specified name dir_name
    // all the files and enclosed directories are also deleted
    public boolean rmdir(String dir_name)
    {
	try {
	    File file = new File(dir_name);
	    String[] files = file.list(); // get list of the files for dir_name
	    for(int i = 0; i < files.length; i++) {
		File aFile = new File(dir_name + File.separator + files[i]);
		if(aFile.isDirectory()) // if directory is found - call rmdir
		    rmdir(dir_name + File.separator + files[i]);
		else
		    aFile.delete(); // delete file
	    }
	    file.delete(); // delete directory
	    return true;
	} catch (Exception ex) {
	    error(ex.getMessage());
	    return false;
	}
    }

    // renames source file or directory to the dest
    public boolean rename(String source, String dest)
    {
	try {
	    File file = new File(source);
	    file.renameTo(new File(dest));
	    return true;
	} catch (Exception ex) {
	    error(ex.getMessage());
	    return false;
	}
    }

    // copies file source to the dest
    protected void copy_file(String source, String dest)  throws FileException
    {
	delete(dest); // if there is file dest - delete it first
	RandomAccessFile source_file, dest_file;
	try {
	    source_file = new RandomAccessFile(source, "r");
	    dest_file = new RandomAccessFile(dest, "rw");
	    // create temp buffer 
	    byte buffer[] = new byte[block_size];      
	    try {
		for(;;) { // read and write data untill eof is not reached
		    int icount = source_file.read(buffer);//read data
		    if(icount == -1)//if end of file is reached
			throw new EOFException();
		    dest_file.write(buffer, 0, icount);//write data
		}
	    }
	    catch(java.io.EOFException ex) {}
	    catch(java.io.IOException ex) {
		source_file.close();
		dest_file.close();
		throw new FileException("File: " + ex.getMessage());
	    }
	    source_file.close(); // close files
	    dest_file.close();
	} catch(java.io.IOException ex) {
	    throw new FileException("File: " + ex.getMessage());
	};
    }

    // copies directory source to the directory dest recursivly
    protected void copy_dir(String source, String dest)  throws FileException
    {
	File sfile = new File(source);
	File dfile = new File(dest);
	String[] files = sfile.list(); // get list of the files for directory
	for(int i = 0; i < files.length; i++) {
	    File aFile = new File(source + File.separator + files[i]);
	    if(aFile.isDirectory()) { // if there is a directory to be copied..
		mkdir(dest + File.separator + files[i]); // create directory
		// call copy_dir recursivly for nested directory
		copy_dir(source + File.separator + files[i], dest + File.separator + files[i]);
	    }
	    else // copy file
		copy_file(source + File.separator + files[i], dest + File.separator + files[i]);
	}
    }

    // copies file to directory, file to file or directory to directory
    public boolean copy(String source, String dest)
    {
	try {
	    File sfile = new File(source);
	    File dfile = new File(dest);
	    if(sfile.isFile()) { // if source is a file
		if(dfile.isDirectory()) // if dest is a directory
		    // copy file to the dest directory
		    copy_file(source, dest + File.separator + source.substring(source.lastIndexOf(File.separatorChar) + 1, source.length()) ); 
		else // copy file to the file with name dest
		    copy_file(source, dest);
	    }
	    else // if source not a file than copy source directory to the dest
		copy_dir(source, dest);
	    return true;
	} catch (Exception ex) {
	    error(ex.getMessage());
	    return false;
	}
    }

    public boolean exists(String file_name)
    {
	try {
	    File file = new File(file_name);
	    return file.exists();
	} catch (Exception ex) {
	    return false;
	}
    }

    // this function is just for testing purposes
    public void test()
    {
	System.out.println("RemoteFile is Ok");
    }

    private void error(String msg)
    {
	System.err.println("**ERROR in RemoteFile: " + msg);
    }
}
